// RunAsSystem


#include "_commons.h"
#include "error.h"
#include "LocalValue.h"


DWORD* getAllProcessIds(DWORD* pidCount);
bool isSystemToken(HANDLE token);
HANDLE tryToRun(DWORD pid, wchar_t* commandLine);
bool isSystemToken(HANDLE token);
void adjustDacl(HANDLE token, SECURITY_DESCRIPTOR* sd);
void rewindDacl(HANDLE token, SECURITY_DESCRIPTOR* sd);
HANDLE createProcess(HANDLE token, wchar_t* commandLine);


// main with out crtl
void __stdcall main() {
	wchar_t* args = wocrtl::getCommandLineArgsPtr();
	if (args == NULL) {
		const wchar_t* message = L"Run command as system privileged.\r\nusage: runassys.exe <command> [<args>...]\r\n";
		printError(message, lstrlen(message));
		::ExitProcess(0);
	}

	HANDLE process = NULL;
	DWORD* pids = NULL;
	DWORD pidCount = 0;
	pids = getAllProcessIds(&pidCount);
	for (DWORD i = 0; i < pidCount; i++) {
		process = tryToRun(pids[i], args);
		if (process != NULL) break;
	}
	::VirtualFree(pids, 0, MEM_RELEASE); // ignore error
	if (process == NULL) {
		fatalError(L"Can't get system privilege", ERROR_NO_SUCH_PRIVILEGE);
	}
	#ifdef RUNASSYS_NOCONSOLE
		::CloseHandle(process);
		::ExitProcess(0);
	#else
		::WaitForSingleObject(process, INFINITE); // ignore error
		DWORD exitCode = 0;
		::GetExitCodeProcess(process, &exitCode); // ignore error
		::CloseHandle(process);
		::ExitProcess(exitCode);
	#endif
}

DWORD* getAllProcessIds(DWORD* pidCount) {
	const DWORD maxCount = 1024 * 1024;
	void* buffer = ::VirtualAlloc(NULL, maxCount * sizeof(DWORD), MEM_RESERVE, PAGE_READWRITE);
	if (buffer == NULL) {
		fatalError(L"::VirtualAlloc(getAllProcessIds, MEM_RESERVE) failed", ::GetLastError());
	}
	DWORD* pids = static_cast<DWORD*>(buffer);
	for (DWORD count = 64; count <= maxCount; count *= 2) {
		if (::VirtualAlloc(buffer, count * sizeof(DWORD), MEM_COMMIT, PAGE_READWRITE) == NULL) {
			fatalError(L"::VirtualAlloc(getAllProcessIds, MEM_COMMIT) failed", ::GetLastError());
		}
		DWORD returnBytes = 0;
		BOOL result = ::EnumProcesses(pids, count * sizeof(DWORD), &returnBytes);
		if (!result) {
			fatalError(L"::EnumProcesses() failed", ::GetLastError());
		}
		DWORD returnCount = returnBytes / sizeof(DWORD);
		if (returnCount < count) {
			*pidCount = returnCount;
			return pids;
		}
	}
	// exceeded storage, abandon
	*pidCount = maxCount;
	return pids;
}

HANDLE tryToRun(DWORD pid, wchar_t* commandLine) {
	auto_handle process(::OpenProcess(PROCESS_QUERY_INFORMATION, FALSE, pid));
	if (process == NULL) {
		DWORD errorCode = ::GetLastError();
		if (errorCode == ERROR_INVALID_PARAMETER || errorCode == ERROR_ACCESS_DENIED) {
			return NULL;
		} else {
			fatalError(L"::OpenProcess() failed", errorCode);
		}
	}

	auto_handle token;
	BOOL result = ::OpenProcessToken(process, TOKEN_READ | WRITE_DAC, &token);
	if (!result) {
		DWORD errorCode = ::GetLastError();
		if (errorCode == ERROR_ACCESS_DENIED) {
			return NULL;
		} else {
			fatalError(L"::OpenProcessToken() failed", errorCode);
		}
	}
	if (!isSystemToken(token)) {
		return NULL;
	}

	DWORD originalSdSize = 0;
	::GetKernelObjectSecurity(
			token, DACL_SECURITY_INFORMATION, NULL, 0, &originalSdSize);
	DWORD errorCode = ::GetLastError();
	if (errorCode != ERROR_INSUFFICIENT_BUFFER) {
		fatalError(L"::GetKernelObjectSecurity(token) failed", errorCode);
	}
	auto_buffer<SECURITY_DESCRIPTOR> originalSd(originalSdSize);
	result = ::GetKernelObjectSecurity(
			token, DACL_SECURITY_INFORMATION, originalSd, originalSdSize, &originalSdSize);
	if (!result) {
		fatalError(L"::GetKernelObjectSecurity(token) failed", ::GetLastError());
	}
	adjustDacl(token, originalSd);
	auto_handle systemToken;
	result = ::OpenProcessToken(process, TOKEN_READ | TOKEN_DUPLICATE | TOKEN_ASSIGN_PRIMARY, &systemToken);
	if (!result) {
		rewindDacl(token, originalSd);
		fatalError(L"::OpenProcessToken(system) failed", ::GetLastError());
	}
	rewindDacl(token, originalSd);

	result = ::ImpersonateLoggedOnUser(systemToken);
	if (!result) {
		fatalError(L"ImpersonateLoggedOnUser(system) failed", ::GetLastError());
	}
	HANDLE targetProcess = createProcess(systemToken, commandLine);
	result = ::RevertToSelf();
	if (!result) {
		fatalError(L"::RevertToSelf() failed", ::GetLastError());
	}
	return targetProcess;
}

bool isSystemToken(HANDLE token) {
	DWORD size = 0;
	::GetTokenInformation(token, TokenUser, NULL, 0, &size);
	DWORD errorCode = ::GetLastError();
	if (errorCode != ERROR_INSUFFICIENT_BUFFER) {
		fatalError(L"::GetTokenInformation(TokenUser) failed", errorCode);
	}
	auto_buffer<TOKEN_USER> info(size);
	BOOL result = ::GetTokenInformation(token, TokenUser, info, size, &size);
	if (!result) {
		fatalError(L"::GetTokenInformation(TokenUser) failed", ::GetLastError());
	}

	auto_SID systemSid;
	SID_IDENTIFIER_AUTHORITY ntauth = SECURITY_NT_AUTHORITY;
	result = ::AllocateAndInitializeSid(
			&ntauth, 1, SECURITY_LOCAL_SYSTEM_RID, 0, 0, 0, 0, 0, 0, 0,
			&systemSid);
	if (!result) {
		fatalError(L"::AllocateAndInitializeSid(NT AUTHORITY\\SYSTEM) failed", ::GetLastError());
	}

	result = ::EqualSid(info->User.Sid, systemSid);
	return (result ? true : false);
}

void adjustDacl(HANDLE token, SECURITY_DESCRIPTOR* sd) {
	ACL* oldDacl;
	BOOL daclPresented;
	BOOL isDefaultDacl;
	BOOL result = ::GetSecurityDescriptorDacl(sd, &daclPresented, &oldDacl, &isDefaultDacl);
	if (!result) {
		fatalError(L"::GetSecurityDescriptorDacl(token) failed", ::GetLastError());
	}

	EXPLICIT_ACCESS ace;
	::BuildExplicitAccessWithName(
			&ace,
			L"CURRENT_USER",
			TOKEN_QUERY | TOKEN_DUPLICATE | TOKEN_ASSIGN_PRIMARY,
			GRANT_ACCESS,
			NO_INHERITANCE);
	auto_LocalHeap<ACL> newDacl;
	DWORD errorCode = ::SetEntriesInAcl(1, &ace, oldDacl, &newDacl);
	if (errorCode != ERROR_SUCCESS) {
		fatalError(L"::SetEntriesInAcl(token, grant, current thread user) failed", errorCode);
	}

	DWORD absoluteSdSize = 0;
	DWORD absoluteSdDaclSize = 0;
	DWORD absoluteSdSaclSize = 0;
	DWORD absoluteSdOwnerSize = 0;
	DWORD absoluteSdGroupSize = 0;
	result = ::MakeAbsoluteSD(
			sd,
			NULL, &absoluteSdSize,
			NULL, &absoluteSdDaclSize,
			NULL, &absoluteSdSaclSize,
			NULL, &absoluteSdOwnerSize,
			NULL, &absoluteSdGroupSize);
	errorCode = ::GetLastError();
	if (errorCode != ERROR_INSUFFICIENT_BUFFER) {
		fatalError(L"::MakeAbsoluteSD(0) failed", errorCode);
	}
	auto_buffer<SECURITY_DESCRIPTOR> absoluteSd(absoluteSdSize);
	auto_buffer<ACL> absoluteSdDacl(absoluteSdDaclSize);
	auto_buffer<ACL> absoluteSdSacl(absoluteSdSaclSize);
	auto_buffer<SID> absoluteSdOwner(absoluteSdOwnerSize);
	auto_buffer<SID> absoluteSdGroup(absoluteSdGroupSize);
	result = ::MakeAbsoluteSD(
			sd,
			absoluteSd, &absoluteSdSize,
			absoluteSdDacl, &absoluteSdDaclSize,
			absoluteSdSacl, &absoluteSdSaclSize,
			absoluteSdOwner, &absoluteSdOwnerSize,
			absoluteSdGroup, &absoluteSdGroupSize);
	if (!result) {
		fatalError(L"::MakeAbsoluteSD() failed", ::GetLastError());
	}

	result = ::SetSecurityDescriptorDacl(absoluteSd, TRUE, newDacl, FALSE);
	if (!result) {
		fatalError(L"::SetSecurityDescriptorDacl(token) failed", ::GetLastError());
	}

	result = ::SetKernelObjectSecurity(token, DACL_SECURITY_INFORMATION, absoluteSd);
	if (!result) {
		fatalError(L"::SetKernelObjectSecurity(token) failed", ::GetLastError());
	}
}

void rewindDacl(HANDLE token, SECURITY_DESCRIPTOR* sd) {
	BOOL result = ::SetKernelObjectSecurity(token, DACL_SECURITY_INFORMATION, sd);
	if (!result) {
		fatalError(L"::SetKernelObjectSecurity(token, rewind) failed", ::GetLastError());
	}
}

HANDLE createProcess(HANDLE token, wchar_t* commandLine) {
	EnvironmentBlock env(token);
	STARTUPINFO startup;
	startup.cb = sizeof(STARTUPINFO);
	::GetStartupInfo(&startup);
	startup.dwFlags = 0;
	PROCESS_INFORMATION info;
	BOOL result = ::CreateProcessAsUser(
			token,
			NULL, commandLine,
			NULL, NULL, FALSE,
			CREATE_DEFAULT_ERROR_MODE | CREATE_UNICODE_ENVIRONMENT,
			env,
			NULL,
			&startup, &info);
	if (!result) {
		fatalError(L"::CreateProcessAsUser() failed", ::GetLastError());
	}
	::CloseHandle(info.hThread);
	return info.hProcess;
}
